# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import pathlib
import typing

import cv2
import lightning.pytorch
import numpy
import numpy.typing
import torch
import torchvision.ops

from tqdm import tqdm

from ...config.data.tbx11k.datamodule import BoundingBox, BoundingBoxes

logger = logging.getLogger(__name__)


def _ordered_connected_components(
    saliency_map: typing.Sequence[typing.Sequence[float]]
    | numpy.typing.NDArray[numpy.double],
) -> list[numpy.typing.NDArray[numpy.bool_]]:
    """Calculates the largest connected components available on a saliency map
    and return those as individual masks.

    This implementation is based on [SCORECAM-2020]_:

    1. Thresholding: The pixel values above 20% of max value are kept in the
       original saliency map.  Everything else is set to zero.
    2. The thresholded saliency map is transformed into a boolean array (ones
       are attributed to all elements above the threshold.
    3. We call :py:func:`skimage.metrics.label` to evaluate all connected
       components and label those with distinct integers.
    4. We histogram the labels and return one binary mask for each label,
       sorted by decreasing size.


    Parameters
    ----------
    saliency_map
        Input saliciency map whose connected components will be calculated
        from.


    Returns
    -------
        A list of boolean masks, one for each connected component, ordered by
        decreasing size.  This list may be empty if the input ``saliency_map``
        is all zeroes.
    """

    # thresholds like [SCORECAM-2020]_
    thresholded_mask = (saliency_map >= (0.2 * numpy.max(saliency_map))).astype(
        numpy.uint8
    )

    # avoids an all zeroes mask being processed
    if not numpy.any(thresholded_mask):
        return []

    # opencv implementation:
    n, labelled = cv2.connectedComponents(thresholded_mask, connectivity=8)
    retval = [labelled == k for k in range(1, n)]

    # scikit-image implementation
    # import skimage.measure
    # labelled, n = skimage.measure.label(thresholded_mask, return_num=True)
    # retval = [labelled == k for k in range(1, n+1)]

    return sorted(retval, key=lambda x: x.sum(), reverse=True)


def _extract_bounding_box(
    mask: numpy.typing.NDArray[numpy.bool_],
) -> BoundingBox:
    """Defines a bounding box surrounding a connected component mask.

    Parameters
    ----------
    mask
        The connected component mask from whom extract the bounding box.


    Returns
    -------
        A bounding box
    """
    x, y, x2, y2 = torchvision.ops.masks_to_boxes(torch.tensor(mask)[None, :])[
        0
    ]
    return BoundingBox(-1, int(x), int(y), int(x2 - x + 1), int(y2 - y + 1))


def _compute_max_iou_and_ioda(
    detected_box: BoundingBox,
    gt_bboxes: BoundingBoxes,
) -> tuple[float, float]:
    """Will calculate how much of detected area lies in ground truth boxes.

    If there are multiple gt boxes, the detected area will be calculated
    for each gt box separately and the gt box with the highest
    intersecting part will be used for the calculation.
    """
    detected_area = detected_box.area()
    if detected_area == 0:
        return 0.0, 0.0

    max_intersection = 0
    max_gt_area = 0

    for bbox in gt_bboxes:
        intersection = bbox.intersection(detected_box)
        if intersection > max_intersection:
            max_intersection = intersection
            max_gt_area = bbox.area()

    if max_gt_area == 0 and max_intersection == 0:
        # This case means no intersection was found, even though there are gt boxes
        iou, ioda = 0.0, 0.0

    else:
        iou = max_intersection / (
            detected_area + max_gt_area - max_intersection
        )
        ioda = max_intersection / detected_area

    return iou, ioda


def _compute_simultaneous_iou_and_ioda(
    detected_box: BoundingBox,
    gt_bboxes: BoundingBoxes,
) -> tuple[float, float]:
    """Will calculate how much of detected area lies between ground truth
    boxes.

    This means that if there are multiple gt boxes, the detected area
    will be compared to them simultaneously (and not to each gt box
    separately).
    """

    detected_area = detected_box.area()
    if detected_area == 0:
        return 0, 0

    intersection = sum([k.intersection(detected_box) for k in gt_bboxes])
    total_gt_area = sum([k.area() for k in gt_bboxes])

    iou = intersection / (detected_area + total_gt_area - intersection)
    ioda = intersection / detected_area

    return float(iou), float(ioda)


def _compute_avg_saliency_focus(
    saliency_map: numpy.typing.NDArray[numpy.double],
    gt_mask: numpy.typing.NDArray[numpy.bool_],
) -> float:
    """Integrates the saliency map over the ground-truth boxes and normalizes
    by total bounding-box area.

    This function will integrate (sum) the value of the saliency map over the
    ground-truth bounding boxes and normalize it by the total area covered by
    all ground-truth bounding boxes.


    Parameters
    ----------
        gt_bboxes
            Ground-truth bounding boxes in the format ``(x, y, width,
            height)``.
        gt_mask
            Ground-truth mask containing the bounding boxes of the ground-truth
            drawn as ``True`` values.


    Returns
    -------
        A single floating-point number representing the Average saliency focus.
    """

    area = gt_mask.sum()
    if area == 0:
        return 0.0

    return numpy.sum(saliency_map * gt_mask) / area


def _compute_proportional_energy(
    saliency_map: numpy.typing.NDArray[numpy.double],
    gt_mask: numpy.typing.NDArray[numpy.bool_],
) -> float:
    """Calculates how much activation lies within the ground truth boxes
    compared to the total sum of the activations (integral).

    Parameters
    ----------
        saliency_map
            A real-valued saliency-map that conveys regions used for
            classification in the original sample.
        gt_mask
            Ground-truth mask containing the bounding boxes of the ground-truth
            drawn as ``True`` values.


    Returns
    -------
        A single floating-point number representing the proportional energy.
    """

    denominator = numpy.sum(saliency_map)

    if denominator == 0.0:
        return 0.0

    return float(numpy.sum(saliency_map * gt_mask) / denominator)


def _process_sample(
    gt_bboxes: BoundingBoxes,
    saliency_map: numpy.typing.NDArray[numpy.double],
) -> tuple[float, float, float, float, BoundingBox]:
    """Calculates the metrics for a single sample.

    Parameters
    ----------
    gt_bboxes
        A list of ground-truth bounding boxes.


    Returns
    -------
        A tuple containing the following values:

        * IoU
        * IoDA
        * Proportional energy
        * Average saliency focus
        * Largest detected bounding box
    """

    masks = _ordered_connected_components(saliency_map)
    detected_box = BoundingBox(-1, 0, 0, 0, 0)
    if masks:
        # we get only the largest bounding box as of now
        detected_box = _extract_bounding_box(masks[0])

    # Calculate localization metrics
    iou, ioda = _compute_max_iou_and_ioda(detected_box, gt_bboxes)

    # The binary_mask will be ON/True where the gt boxes are located
    binary_mask = numpy.zeros_like(saliency_map, dtype=numpy.bool_)
    for bbox in gt_bboxes:
        binary_mask[
            bbox.ymin : bbox.ymin + bbox.height,
            bbox.xmin : bbox.xmin + bbox.width,
        ] = True

    return (
        iou,
        ioda,
        _compute_proportional_energy(saliency_map, binary_mask),
        _compute_avg_saliency_focus(saliency_map, binary_mask),
        detected_box,
    )


def run(
    input_folder: pathlib.Path,
    datamodule: lightning.pytorch.LightningDataModule,
) -> dict[str, list[typing.Any]]:
    """Applies visualization techniques on input CXR, outputs images with
    overlaid heatmaps and csv files with measurements.

    Parameters
    ---------
    input_folder
        Directory in which the saliency maps are stored for a specific
        visualization type.
    datamodule
        The lightning datamodule to iterate on.


    Returns
    -------

        A dictionary where keys are dataset names in the provide datamodule,
        and values are lists containing sample information alongside metrics
        calculated:

        * Sample name (str)
        * Sample target class (int)
        * IoU (float)
        * IoDA (float)
        * Proportional energy (float)
        * Average saliency focus (float)
        * Largest detected bounding box (x, y, width, height) (4 x int)
    """

    retval: dict[str, list[typing.Any]] = {}

    # TODO: This loads the images from the dataset, but they are not useful at
    # this point.  Possibly using the contents of ``datamodule.splits`` can
    # substantially speed this up.
    for dataset_name, dataset_loader in datamodule.predict_dataloader().items():
        logger.info(
            f"Estimating interpretability metrics for dataset `{dataset_name}`..."
        )
        retval[dataset_name] = []

        for sample in tqdm(
            dataset_loader, desc="batches", leave=False, disable=None
        ):
            name = str(sample[1]["name"][0])
            label = int(sample[1]["label"].item())

            # TODO: This is very specific to the TBX11k system for labelling
            # regions of interest.  We need to abstract from this to support more
            # datasets and other ways to annotate.
            bboxes: BoundingBoxes = sample[1].get(
                "bounding_boxes", BoundingBoxes()
            )

            if label == 0:
                # we add the entry for dataset completeness
                retval[dataset_name].append([name, label])
                continue

            if not bboxes:
                logger.warning(
                    f"Sample `{name}` does not contdain bounding-box information. "
                    f"No localization metrics can be calculated in this case. "
                    f"Skipping..."
                )
                # we add the entry for dataset completeness
                retval[dataset_name].append([name, label])
                continue

            # we fully process this entry
            retval[dataset_name].append(
                [
                    name,
                    label,
                    *_process_sample(
                        bboxes,
                        numpy.load(
                            input_folder
                            / pathlib.Path(name).with_suffix(".npy")
                        ),
                    ),
                ]
            )

    return retval
