# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os

import click

from clapper.click import verbosity_option
from clapper.logging import setup

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


def _sorting_rule(folder_name):
    folder_name = os.path.basename(folder_name)

    if "gradcam" == folder_name:
        return (0, folder_name)
    elif "scorecam" == folder_name:
        return (1, folder_name)
    elif "fullgrad" == folder_name:
        return (2, folder_name)
    elif "randomcam" == folder_name:
        return (4, folder_name)
    else:  # Everything else will be sorted alphabetically after fullgrad and before randomcam
        return (3, folder_name)


def _get_images_from_directory(dir_path):
    image_files = []
    for root, _, files in os.walk(dir_path):
        for file in files:
            if file.lower().endswith((".png", ".jpg", ".jpeg")):
                image_files.append(os.path.join(root, file))
    return image_files


@click.command(
    epilog="""Examples:

\b
    1. Compares different visualization techniques by showing their results side-by-side
    (on a per image basis). The input_folder should be a parent folder containing different
    folders for each visualization technique (e.g. input_folder/gradcam, input_folder/scorecam, etc.):

       .. code:: sh

          ptbench compare-vis -i path/to/input_folder -o path/to/output_folder
""",
)
@click.option(
    "--input-folder",
    "-i",
    help="Path to the parent folder of resulted visualizations from visualize command",
    required=True,
    type=click.Path(),
)
@click.option(
    "--output-folder",
    "-o",
    help="Path where to store the cam comparisons (created if does not exist)",
    required=True,
    default="cam_comparisons",
    type=click.Path(),
)
@verbosity_option(logger=logger, expose_value=False)
def compare_vis(input_folder, output_folder) -> None:
    """Compares multiple visualization techniques by showing their results in
    one image."""

    import os

    import cv2
    import matplotlib.pyplot as plt

    assert (
        input_folder != output_folder
    ), "Output folder must not be the same as the input folder."

    # Check that the output folder is not a subdirectory of the input_folder
    assert not output_folder.startswith(
        input_folder
    ), "Output folder must not be a subdirectory of the input folder."

    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    dir_element_names = os.listdir(input_folder)

    # Sort folders by visualization type
    dir_element_names.sort(key=_sorting_rule)

    # Get the list of subfolders
    vis_type_folders = [
        os.path.join(input_folder, d)
        for d in dir_element_names
        if os.path.isdir(os.path.join(input_folder, d))
    ]

    if not vis_type_folders:
        raise ValueError("No subdirectories found in the parent folder.")

    for vis_type_folder in vis_type_folders:
        target_class_folders = next(os.walk(vis_type_folder))[1]
        for target_class_folder in target_class_folders:
            current_target_class_dir = os.path.join(
                vis_type_folder, target_class_folder
            )
            dataset_type_folders = next(os.walk(current_target_class_dir))[1]
            for dataset_type_folder in dataset_type_folders:
                comparison_folders = [
                    os.path.join(
                        input_folder,
                        vis_type,
                        target_class_folder,
                        dataset_type_folder,
                    )
                    for vis_type in dir_element_names
                ]

                valid_folders = []

                for folder in comparison_folders:
                    if not os.path.exists(folder):
                        logger.warning(f"Folder does not exist: {folder}")
                    else:
                        valid_folders.append(folder)

                comparison_folders = valid_folders

                output_directory = os.path.join(
                    output_folder, target_class_folder, dataset_type_folder
                )
                os.makedirs(output_directory, exist_ok=True)

                # Use a set (unordered collection of unique elements) for efficient membership tests
                image_names = {
                    os.path.basename(img)
                    for img in _get_images_from_directory(comparison_folders[0])
                }

                # Only keep image names that exist in all folders
                for folder in comparison_folders[1:]:
                    # This is basically an intersection-check of contents of different folders
                    # Images that don't exist in all folders are removed from the set
                    image_names &= {
                        os.path.basename(img)
                        for img in _get_images_from_directory(folder)
                    }

                if not image_names:
                    raise ValueError("No common images found in the folders.")

                max_cols = min(5, len(comparison_folders))
                rows, cols = (
                    len(comparison_folders) // max_cols,
                    len(comparison_folders) % max_cols,
                )
                rows = rows if cols == 0 else rows + 1

                for image_name in image_names:
                    fig, axs = plt.subplots(
                        rows, max_cols, figsize=(4 * max_cols, 4 * rows)
                    )
                    fig.subplots_adjust(
                        left=0,
                        right=1,
                        bottom=0,
                        top=1,
                        wspace=0.05,
                        hspace=0.1,
                    )
                    axs = axs.ravel()
                    for i, folder in enumerate(comparison_folders):
                        image_path = [
                            img
                            for img in _get_images_from_directory(folder)
                            if os.path.basename(img) == image_name
                        ][0]
                        try:
                            img = cv2.imread(image_path)
                            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                        except cv2.error as e:
                            raise RuntimeError(
                                "Error reading or processing image: "
                                + image_path
                            ) from e
                        axs[i].imshow(img)
                        axs[i].set_title(
                            os.path.basename(
                                os.path.dirname(os.path.dirname(folder))
                            ),
                            fontsize=20,
                        )
                        axs[i].axis("off")

                    # Fill the remaining columns with empty plots (white)
                    for i in range(len(comparison_folders), rows * max_cols):
                        axs[i].axis("off")

                    plt.savefig(
                        os.path.join(output_directory, f"{image_name}"),
                        bbox_inches="tight",
                        pad_inches=0,
                    )
                    plt.close(fig)
