# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib

import click

from clapper.click import ResourceOption, verbosity_option
from clapper.logging import setup

from .click import ConfigCommand

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


@click.command(
    entry_point_group="ptbench.config",
    cls=ConfigCommand,
    epilog="""Examples:

1. Evaluate the generated saliency maps for their localization performance:

   .. code:: sh

      ptbench saliency-interpretability -vv tbx11k-v1-healthy-vs-atb --input-folder=parent_folder/gradcam/ --output-json=parent_folder/gradcam/tbx11k-v1-interp.json

""",
)
@click.option(
    "--datamodule",
    "-d",
    help="""A lighting data module that will be asked for prediction data
    loaders. Typically, this includes all configured splits in a datamodule,
    however this is not a requirement.  A datamodule that returns a single
    dataloader for prediction (wrapped in a dictionary) is acceptable.""",
    required=True,
    cls=ResourceOption,
)
@click.option(
    "--input-folder",
    "-i",
    help="""Path where to load saliency maps from.  You can generate saliency
    maps with ``ptbench generate-saliencymaps``.""",
    required=True,
    type=click.Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
        path_type=pathlib.Path,
    ),
    default="saliency-maps",
    cls=ResourceOption,
)
@click.option(
    "--output-json",
    "-o",
    help="""Path where to store the output JSON file containing all
    measures.""",
    required=True,
    type=click.Path(
        file_okay=True,
        dir_okay=False,
        path_type=pathlib.Path,
    ),
    default="saliency-interpretability.json",
    cls=ResourceOption,
)
@verbosity_option(logger=logger, cls=ResourceOption, expose_value=False)
def saliency_interpretability(
    datamodule,
    input_folder,
    output_json,
    **_,
) -> None:
    """Evaluates saliency map agreement with annotations (human
    interpretability).

    The evaluation happens by comparing saliency maps with ground-truth
    provided by any other means (typically following a manual annotation
    procedure).

    .. note::

       For obvious reasons, this evaluation is limited to databases that
       contain built-in annotations which corroborate classification.


    As a result of the evaluation, this application creates a single JSON file
    that resembles the original datamodule, with added information containing
    the following measures, for each sample:

    * IoU: The intersection of the (thresholded) saliency maps with
      the annotation the most overlaps, over the union of both areas.
    * IoDA: The intersection of the (thresholded) saliency maps with
      the annotation that most overlaps, over area of (thresholded) saliency
      maps.
    * Proportional Energy: A measure that compares (UNthresholed) saliency maps
      with annotations (based on [SCORECAM-2020]_). It estimates how much
      activation lies within the ground truth boxes compared to the total sum
      of the activations.
    * Average Saliency Focus: estimates how much of the ground truth bounding
      boxes area is covered by the activations.  It is similar to the
      proportional energy measure in the sense it does not need explicit
      thresholding.

    .. important::

       The thresholding algorithm used to evaluate IoU and IoDA measures is
       based on the process done by the original CAM paper [GRADCAM-2015]_. It
       keeps all points from the saliency map that are above the 20% of its
       maximum value.

       It then calculates a **single** bounding box for largest connected
       component.  This bounding box represents detected elements on the
       original sample that corroborate the classification outcome.

       IoU and IoDA are only evaluated for a single ground-truth bounding box
       per sample (the one with the highest overlap).  Any other bounding box
       marked on the sample is ignored in the present implementation.
    """

    import json

    from ..engine.saliency.interpretability import run

    datamodule.model_transforms = []
    datamodule.prepare_data()
    datamodule.setup(stage="predict")

    results = run(input_folder, datamodule)

    with output_json.open("w") as f:
        logger.info(f"Saving output file to `{str(output_json)}`...")
        json.dump(results, f, indent=2)
