# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import glob
import os
import shutil

import cv2
import numpy as np
import pandas as pd


def _calculate_stats_over_dataset(file_path):
    data = pd.read_csv(file_path)

    # Compute mean, median and standard deviation for each score type
    metrics = {}
    for column in [
        "MoRF",
        "LeRF",
        "Combined Score ((LeRF-MoRF) / 2)",
        "IoU",
        "IoDA",
        "propEnergy",
        "ASF",
    ]:
        mean = round(data[column].mean(), 3)
        median = round(data[column].median(), 3)
        std_dev = round(data[column].std(), 3)
        q1 = round(data[column].quantile(0.25), 3)
        q3 = round(data[column].quantile(0.75), 3)

        metrics[column] = {
            "mean": mean,
            "median": median,
            "std_dev": std_dev,
            "Q1": q1,
            "Q3": q3,
        }

    return metrics


def calculate_metrics_avg_for_every_class(input_folder):
    directories = [
        dir
        for dir in os.listdir(input_folder)
        if os.path.isdir(os.path.join(input_folder, dir))
    ]

    # Get unique file names ending with .csv without the extension
    csv_file_names = set()
    for directory in directories:
        dir_path = os.path.join(input_folder, directory)
        csv_files_in_dir = glob.glob(os.path.join(dir_path, "*.csv"))
        for csv_file in csv_files_in_dir:
            csv_file_name = os.path.splitext(os.path.basename(csv_file))[0]
            csv_file_names.add(csv_file_name)

    all_results = {file_name: [] for file_name in csv_file_names}

    for file_name in csv_file_names:
        previous_summary = os.path.join(
            input_folder, f"{file_name}_summary.csv"
        )
        if os.path.exists(previous_summary):
            backup = previous_summary + "~"
            if os.path.exists(backup):
                os.unlink(backup)
            shutil.move(previous_summary, backup)

    for directory in directories:
        dir_path = os.path.join(input_folder, directory)

        for file_name in csv_file_names:
            csv_file = os.path.join(dir_path, f"{file_name}.csv")

            if os.path.exists(csv_file):
                class_name = directory
                metrics = _calculate_stats_over_dataset(csv_file)

                for score_type, values in metrics.items():
                    result = {
                        "Class Name": class_name,
                        "Score Type": score_type,
                        "Mean": values["mean"],
                        "Standard Deviation": values["std_dev"],
                        "Median": values["median"],
                        "Q1": values["Q1"],
                        "Q3": values["Q3"],
                    }
                    all_results[file_name].append(result)

    for file_name, results in all_results.items():
        df_results = pd.DataFrame(results)
        df_results.to_csv(
            os.path.join(input_folder, f"{file_name}_summary.csv"), index=False
        )


def draw_boxes_on_image(img_with_boxes, bboxes):
    for i, bbox in enumerate(bboxes):
        xmin, ymin = int(bbox[1].item()), int(bbox[2].item())
        width, height = int(bbox[3].item()), int(bbox[4].item())
        xmax = int(xmin + width)
        ymax = int(ymin + height)
        cv2.rectangle(
            img_with_boxes, (xmin, ymin), (xmax, ymax), (0, 255, 0), 2
        )

        # Add text to the bbox
        text = f"GTruth bbox {i+1}"
        font_scale = max(
            0.5, min(width, height) / 200.0
        )  # Ensure minimum font scale of 0.5
        font_scale = min(font_scale, 1.3)  # Ensure maximum font scale of 1.3

        # Computes text height
        (text_width, text_height), baseline = cv2.getTextSize(
            text, cv2.FONT_HERSHEY_SIMPLEX, font_scale, 2
        )

        # Make sure not to draw text outside of image
        if ymin + height + text_height + 2 < img_with_boxes.shape[0]:
            y = ymin + height + text_height + 2  # put below box
        else:
            y = ymin - text_height - 2  # put above box

        y = int(y)

        cv2.putText(
            img_with_boxes,
            text,
            (xmin + 5, y),
            cv2.FONT_HERSHEY_SIMPLEX,
            font_scale,
            (36, 255, 12),
            2,
        )
    return img_with_boxes


def draw_largest_component_bbox_on_image(img_with_boxes, x, y, w, h):
    cv2.rectangle(img_with_boxes, (x, y), (x + w, y + h), (0, 0, 255), 2)

    # Add text to the bbox
    text = "Detected Area"
    font_scale = max(0.5, min(w, h) / 200.0)  # Ensure minimum font scale of 0.5
    font_scale = min(font_scale, 1.3)  # Ensure maximum font scale of 1.3

    # Computes text height
    (text_width, text_height), baseline = cv2.getTextSize(
        text, cv2.FONT_HERSHEY_SIMPLEX, font_scale, 2
    )

    # Make sure not to draw text outside bottom of image
    if y + h + text_height + 2 < img_with_boxes.shape[0]:
        y = y + h + text_height + 2  # put below box
    else:
        y = y - text_height - 2  # put above box

    y = int(y)

    cv2.putText(
        img_with_boxes,
        text,
        (x + 5, y),
        cv2.FONT_HERSHEY_SIMPLEX,
        font_scale,
        (0, 0, 255),
        2,
    )
    return img_with_boxes


def show_cam_on_image(
    img: np.ndarray,
    thresholded_mask: np.ndarray,
    use_rgb: bool = False,
    colormap: int = cv2.COLORMAP_JET,
    image_weight: float = 0.5,
) -> np.ndarray:
    """This function overlays the cam mask on the image as an heatmap. By
    default the heatmap is in BGR format.

    :param img: The base image in RGB or BGR format.
    :param mask: The cam mask.
    :param threshold: Only apply heatmap to areas where cam is above this threshold.
    :param use_rgb: Whether to use an RGB or BGR heatmap, this should be set to True if 'img' is in RGB format.
    :param colormap: The OpenCV colormap to be used.
    :param image_weight: The final result is image_weight * img + (1-image_weight) * mask.
    :returns: The default image with the cam overlay.

    This is a slightly modified version of the show_cam_on_image implementation in:
    https://github.com/jacobgil/pytorch-grad-cam
    """

    if img.shape[:2] != thresholded_mask.shape:
        raise ValueError(
            "The shape of the mask should be the same as the shape of the image."
        )

    heatmap = cv2.applyColorMap(np.uint8(255 * thresholded_mask), colormap)
    if use_rgb:
        heatmap = cv2.cvtColor(heatmap, cv2.COLOR_BGR2RGB)
    heatmap = np.float32(heatmap) / 255

    if np.max(img) > 1:
        raise Exception("The input image should np.float32 in the range [0, 1]")

    if image_weight < 0 or image_weight > 1:
        raise Exception(
            f"image_weight should be in the range [0, 1].\
                Got: {image_weight}"
        )

    # For pixels where the mask is zero,
    # the original image pixels are being used without a mask.
    cam = np.where(
        thresholded_mask[..., np.newaxis] == 0,
        img,
        (1 - image_weight) * heatmap + image_weight * img,
    )
    cam = cam / np.max(cam)
    return np.uint8(255 * cam)


def visualize_road_scores(
    visualization, MoRF_score, LeRF_score, combined_score, name, percentiles
):
    visualization = cv2.putText(
        visualization,
        name,
        (10, 20),
        cv2.FONT_HERSHEY_SIMPLEX,
        0.4,
        (255, 255, 255),
        1,
        cv2.LINE_AA,
    )
    visualization = cv2.putText(
        visualization,
        f"Percentiles: {percentiles}",
        (10, 40),
        cv2.FONT_HERSHEY_SIMPLEX,
        0.4,
        (255, 255, 255),
        1,
        cv2.LINE_AA,
    )
    visualization = cv2.putText(
        visualization,
        "Remove and Debias",
        (10, 55),
        cv2.FONT_HERSHEY_SIMPLEX,
        0.4,
        (255, 255, 255),
        1,
        cv2.LINE_AA,
    )
    visualization = cv2.putText(
        visualization,
        f"MoRF score: {MoRF_score:.5f}",
        (10, 70),
        cv2.FONT_HERSHEY_SIMPLEX,
        0.4,
        (255, 255, 255),
        1,
        cv2.LINE_AA,
    )
    visualization = cv2.putText(
        visualization,
        f"LeRF score: {LeRF_score:.5f}",
        (10, 85),
        cv2.FONT_HERSHEY_SIMPLEX,
        0.4,
        (255, 255, 255),
        1,
        cv2.LINE_AA,
    )
    visualization = cv2.putText(
        visualization,
        f"(LeRF-MoRF)/2: {combined_score:.5f}",
        (10, 100),
        cv2.FONT_HERSHEY_SIMPLEX,
        0.4,
        (255, 255, 255),
        1,
        cv2.LINE_AA,
    )
    return visualization
